module TSInterface;
{---------------------------------------------------------
{       Copyright 1981, Three Rivers Computer Corp.
{       by W. J.Hansen
{
{  Abstract:
{       Pascal interface to tape streamer micro code.
{--------------------------------------------------------}

{---------------------------------------------------------
{  Change Log:
{
{       11 Feb 83  V2.8  WJH    Merge WaitForRewound into TSPosn.
{
{        7 Feb 83  V2.7  WJH    Don't TSReset in TSPosn and TSRdEnd.
{                               Add Condition as parameter to SendCmd
{                                 and CallTSCmd.
{
{       25 Jan 83  V2.6  WJH    Define bit four of 2nd status word.
{
{       11 Oct 82  V2.5  WJH    Change TSInit to test whether streamer
{                                 timed out instead of whether it set
{                                 the PerqReSetStreamer bit
{                                 (streamer appears to not indicate ReSet
{                                 if streamer has not had a command from
{                                 one reset to the next)
{
{       20 Nov 81  V2.4  WJH    change EnsureRewound to WaitForRewound
{                                 & put rewind commands before calls
{                               have CallTSPosn call WaitForRewound
{                               indicate write protection in short PrtStatus
{
{        8 Oct 81  V2.3  WJH    fix check for BOT
{
{        7 Oct 81  V2.2  WJH    more careful testing of status in EnsureRewound
{                               allow write-protected as OK in PrtStatus
{
{       15 Sep 81  V2.1  WJH    delay in EnsureRewound after streamer claims
{                                  it is READY
{                               other delays as needed to wait for 
{                                  end of possible tape movement
{
{       11 Sep 81  V2.0  WJH    adapt to removal from TSIO of
{                                  WrSetup, WREof, RdEof, TSPosn
{                               use Wind, Pause, Rewind in TSInit
{                                  to avoid going off front of tape
{                               make sure that Status is stored by TSPosn
{                               no internal routine uses TSCmd
{                               ignore write protect bit
{
{       10 Sep 81  V1.4  WJH    send rewind command in TSInit
{                               make EnsureRewound and calls on it
{                               call TSReset and TSCmd from CallRdEnd
{
{        9 Sep 81  V1.3  WJH    simplify message if reset status is ok
{                               wait for rewind in TSPosn, RdSetup, WrSetup
{
{        8 Sep 81  V1.2  WJH    all TSIO routines are now predicates
{                               use TSResetDrive in TSInit
{                               include CallTSReset
{
{       21 Aug 81  V1.1  WJH    RdBuffers saves status for Rdend
{                               accumulate statistics 
{
{       20 Aug 81  V1.0  WJH    recombine with TSRead and TSWrite
{                               add Msg arg to TSError
{                               WrEnd and RdEnd return Status (read OnLine)
{                               TSInit rewinds the tape
{                               PrtStatus takes a Msg and prints it
{
{ (TsWrite: 14 Aug 81  V0.3  WJH  remove status parm from WrEnd
{ (
{ (         12 Aug 81  v0.2  WJH
{ (                    remove pieces now in TSUtilities
{ (                    add Start arg to WrBuffers
{
{ (TSRead:
{ (         14 Aug 81  V0.3  WJH    remove status parameter from RdEnd
{ (   
{ (         12 aug 81  v0.2  WJH    create tsread from tswrite
{ (                                 remove stuff moved to TSUtilities
{ (                                 add Start parm to RdBuffers
{
{       14 Aug 81  V0.5  WJH    add CallTSTestBusy
{
{       13 Aug 81  V0.4  WJH    add CallTSCmd
{                               make TSPosn not a predicate
{
{       12 Aug 81  v0.3  WJH    create TSUtilities from tsread
{
{       12 aug 81  v0.2  WJH    create tsread from tswrite
{
{       11 aug 81  v0.1  WJH    add WREof and TSCmd
{
{       13 Jul 81  v0.0  WJH
{              Create module tswrite
{
{--------------------------------------------------------}

EXPORTS

imports StBufUts from StBufUts;

type
   StatBuf = array [1..7] of integer;
   TSMsg   = String[64];
   LineSetting = (OffLine, OnLine, NoChange);

procedure  TSInit;
procedure  CallTSPosn(Command: integer; var Status: StatBuf);
procedure  CallTSCmd(Command: integer;  Condition: LineSetting);
function   CallTSTestBusy(var Status: StatBuf): Boolean;
procedure  PrtStatus (var S: StatBuf;  Msg: TSMsg);

procedure  CallWrSetup(var Status: StatBuf);
procedure  CallWrBuffers(BufCnt: integer;var BufLocLoc:BufList; Start:integer);
procedure  CallWrEnd(var Status: StatBuf);
procedure  CallWrEof;

procedure  CallRdSetup(var Status: StatBuf);
function   CallRdBuffers(var BufCnt: integer;
                         var BufLocLoc: BufList; Start: integer): Boolean;
procedure  CallRdEnd(var Status: StatBuf);
procedure  CallRdEof;


exception  TSError(Status: StatBuf;  Msg: TSMsg);
{---------------------------------------------------------
{ Abstract: 
{    This (PASCAL-)exception is raised when any (hardware-)exception
{    other than FileMark is returned from the streamer.  
{ Parameter:
{    S - a StatBuf to hold the Status.
{    Msg - an explanation of the failure.
{--------------------------------------------------------}

procedure  CallTSReset(var Status: StatBuf);


PRIVATE

imports ControlStore from ControlStore;   {for LoadControlStore}
imports SystemDefs   from SystemDefs;     {for 'double'}
imports IO_Others    from IO_Others;      {for IOGetTime}

const
{$Include tsio.dfs}   {def'ns of microstore locations of routines}

var 
   SoftErrors, OverRuns: integer;      {Accumulate statistics}

procedure Pause(Ticks: integer);
{---------------------------------------------------------
{ Abstract:
{    Holds execution for specified number of clock ticks.
{ Parameter:
     Ticks - Number of ticks to wait.  (60 or 50 per second).
{--------------------------------------------------------}
var 
   LastTick, CurTick: double;
begin
   IOGetTime(CurTick);
   LastTick := CurTick;
   while Ticks > 0 do begin      {count down requested number of clock ticks}
      Ticks := Ticks -1;
      while LastTick[0]=CurTick[0] do    {wait for a tick}
         IOGetTime(CurTick);
      LastTick := CurTick;
   end;
end;


procedure SendCmd(Command: integer;  Condition: LineSetting; 
                    var Status: StatBuf; var Success: Boolean);
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine that performs a command.
{ Parameter:
{    Command - the command to be executed.  
{          (Write=#100,  WriteFM=#140,  Read=#200, ReadFM=#240,
{           #041 is rewind, #042 is erase tape, and #044 is retension, 
{           for others, see streamer documentation.)
{    Condition -  0: set streamer offline,  1: set streamer online;
{              other: don't change whether it is offline or on.
{    Status  - If Success is False, this holds the status.
{    Success - Set true if microcode signals success and false otherwise.
{--------------------------------------------------------}
var 
   T: Boolean;
begin
   LoadAdr(Status);
   LoadExpr(Condition);
   LoadExpr(Command);
   LoadExpr(LOR (Shift(TSCmd, 8), Shift(TSCmd, -8)));
   InLineByte(240 {STLATE});   {TLATE the Status address}
   InLineByte(4);              { stack location of address to TLATE }
   InLineByte(#277 {JCS});
   StorExpr(T);
   Success := T;
end;

   
procedure  TSInit;
{---------------------------------------------------------
{  Abstract:
{      Loads tape streamer microcode and rewinds tape.
{  Environment:
{      Assume file "TSIO.BIN" contains tape streamer microcode.
{  Side Effect:
{      Leaves streamer OffLine.
{  Error:
{      TSError(failed to start rewind) if dont get 'Perq reset streamer'.
{  Notes:
{      Status[1] bits are set for any problem in starting streamer.
{      Status[2] has #201 to mean 'Perq reset the streamer';
{      Status[2] bits #210 mean that the tape is at beginning of tape.
{--------------------------------------------------------}
var
   MicroCode: MicroFile;
   Status: StatBuf;
   T: Boolean;
   CR: String;
begin
   Reset (MicroCode, 'TSIO.BIN');
   LoadControlStore(MicroCode);
   Close(MicroCode);
   CallTSReset(Status);   
   while Land(Status[1], #100) = #100 do begin
      writeln('Please insert tape cartridge in drive and press RETURN.');
      readln(CR);
      CallTSReset(Status);   
   end;
   if (Status[1]<>0) and (Status[1]<>#220{write protected})  then
      Raise TSError(Status, 'Failed to reset streamer drive')
   else if Land (Status[2], #2) = #2 then 
      Raise TSError(Status, 'Streamer did not respond');

{ The following code ensures that the tape is not between the last two holes.
{ It fails because the Wind command does a Rewind first.
{
{  if Land(Status[2], #10) <> #10 then begin  {not at BOT}
{     SendCmd(#044, NoChange, Status, T);    {Wind}
{     if not T then
{        Raise TSError(Status, 'Could not start initial wind');
{     Pause(60);  {Let it wind out for one second (at least one hole),
{                  in case it was between start of tape holes.}
{     CallTSReset(Status);
{     Pause(30);            {Let tape stop moving}
{     SendCmd(#041, NoChange, Status, T); {REWIND}
{     if not T then
{        Raise TSError(Status, 'Could not rewind');
{  end;
{}
end;


procedure  CallTSReset(var Status: StatBuf);
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine to reset the streamer.
{ Parameter:
{    Status - The status after resetting streamer is
{        stored in this array.
{--------------------------------------------------------}
var 
   T: Boolean;
begin
   LoadAdr(Status);
   LoadExpr(LOR (Shift(TSReset, 8), Shift(TSReset, -8)));
   InLineByte(228 {TLATE2});
   InLineByte(#277 {JCS});
   StorExpr(T);
end;


procedure  CallTSPosn(Command: integer; var Status: StatBuf);
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine that performs a command.
{ Parameter:
{    Command - the command to be executed.  (#041 is rewind,
{        #042 is erase tape, and #044 is retension, 
{        for others, see streamer documentation.)
{    Status  - set to the status of streamer after resetting streamer
{ Side Effects:
{    Leaves streamer performing the command.
{--------------------------------------------------------}
var 
   T: Boolean;

 function StatusOK: Boolean;
   begin
      StatusOK := ((Status[1]=#0) or (Status[1]=#220{Write protected})) 
         and (Land(Status[2], #166{ignore BOT & Reset}) = #0); 
   end;

begin
   SendCmd(Command, OffLine, Status, T);
   if T then begin {wait for rewind}  {(in fact, T is always true)}
      pause(10);  {wait for command to get settled}
      if CallTSTestBusy(Status) then begin
          write('Waiting for streamer');
          while CallTSTestBusy(Status) do begin
            write('*');
            pause(60);
          end;
          writeln;
          if  (Command<#100) and (Land(Status[2], #10) <> #10) then
             writeln('Stream not at beginning of tape');
          if not StatusOK then 
             raise TSError(Status,
                           'Unexpected status while waiting for streamer');
          Pause(20);  {Wait 1/3 second to be SURE streamer is ready}
      end;
   end;
end;

procedure  CallTSCmd(Command: integer;  Condition: LineSetting);
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine that performs a command.
{ Parameter:
{    Command - the command to be executed.  
{          (Write=#100,  WriteFM=#140,  Read=#200, ReadFM=#240)
{    Condition - Choose OnLine/OffLine/NoChange.
{ Errors:
{    Raises TSError if a streamer exception occurs.
{ Environment:
{    Assumes the streamer has been reset or is in write or read command.
{--------------------------------------------------------}
var
   T: Boolean;
   Status: StatBuf;
begin
   SendCmd(Command, Condition, Status, T);
   if not T then raise TSError(Status, 'TSCmd failed');
end;

function   CallTSTestBusy(var Status: StatBuf): Boolean;
{---------------------------------------------------------
{ Abstract:
{    Calls microcode that test the streamer readiness.
{ Parameter:
{    Status  - When streamer is not busy, CallTSTestBusy returns
{          FALSE and stores streamer status in Status.
{ Returns:
{    TRUE if busy. FALSE if ready or had error.   
{ Side Effects:
{    If returns FALSE, streamer has been left OffLine.
{--------------------------------------------------------}
var
   T: Boolean;
begin
   LoadAdr(Status);
   LoadExpr(LOR (Shift(TSTestBusy, 8), Shift(TSTestBusy, -8)));
   InLineByte(228 {TLATE2});   {TLATE the Status address}
   InLineByte(#277 {JCS});
   StorExpr(T);
   CallTSTestBusy := T;
end;

procedure  PrtStatus (var S: StatBuf;  Msg: TSMsg);
{---------------------------------------------------------
{ Abstract:
{    Prints the status data from the streamer.
{ Parameter:
{    Status - a buffer containing status information.
{          The six status bytes are stored one to a word.
{--------------------------------------------------------}
begin
   if ((S[1]=0) or (S[1]=#220{write prot})) and (S[4]=0) and (S[6]=0)
            and (Land(S[2], #166{ignore BOT and reset}) = 0) then
      if S[1]=#220{wr prot} then
         writeln(Msg, ' - Tape is OK and write protected')
      else
         writeln(Msg, ' - OK (Tape NOT write protected)')
   else begin
      writeln;
      writeln('Streamer Status -   ', Msg);

      if (LAND(S[1], #177)<>#0)  and  (LAND(S[1], #200)=#0) then
         writeln('Streamer hardware error: Exception group 0 bit missing.');
      if (LAND(S[1], #177)=#0)  and  (LAND(S[1], #200)=#1) then
        writeln('Streamer hardware error: Superfluous exception group 0 bit.');
      if LAND(S[1], #100)=#100 then
         writeln('    No cartridge detected in drive.');
      if LAND(S[1], #40)=#40 then
         writeln('    Program error: drive not on-line.');
      if LAND(S[1], #20)=#20 then
         writeln('    Write protected tape.');
      if LAND(S[1], #10)=#10 then
         writeln('    End of tape detected before end of data.');
      if LAND(S[1], #4)=#4 then
         writeln('    Could not recover from tape data error.');
      if LAND(S[1], #2)=#2 then
         writeln('    Couldn''t read a block, couldn''t even find it.');
      if LAND(S[1], #1)=#1 then
         writeln('    File mark detected on tape.');

      if (LAND(S[2], #177)<>#0)  and  (LAND(S[2], #200)=#0) then
         writeln('Streamer hardware error: Exception group 1 bit missing.');
      if (LAND(S[2], #177)=#0)  and  (LAND(S[2], #200)=#1) then
        writeln('Streamer hardware error: Superfluous exception group 1 bit.');
      if LAND(S[2], #100)=#100 then
         writeln('    Program error: undefined command from Perq.');
      if LAND(S[2], #40)=#40 then
         writeln('    No data found on streamer tape.');
      if LAND(S[2], #20)=#20 then
         writeln('    Elderly cartridge: 8 or more read retries for a block.');
      if LAND(S[2], #10)=#10 then
         writeln('    Streamer at beginning of tape.');
      if LAND(S[2], #4)=#4 then
         writeln('       (Undefined bit 2 was set).');
      if LAND(S[2], #1)=#1 then
         writeln('    Perq reset the streamer.');

      if LAND(S[2], #2)=#2 then begin
         writeln('    Streamer not responding.');
         writeln('   Waiting at point', S[4]:6:8);
         writeln('   Last command was', S[3]:6:8);
      end 
      else begin
         writeln(S[3]*256+S[4]:6, ' Soft errors counted');
         writeln(S[5]*256+S[6]:6, ' Speed mismatch count');
      end;
   end;
end;

procedure  CallWrSetup(var Status: StatBuf);
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine to initialize streamer
{    and set up a write command.
{ Side Effects:
{    Streamer is left ONLINE with a Write command.
{ Parameter:
{    Status - The status after intitializing streamer is
{        stored in this array.
{--------------------------------------------------------}
begin
   CallTSPosn(#041{rewind}, Status);
end;

procedure CallWrBuffers(BufCnt: integer; var BufLocLoc:BufList; Start:integer);
{---------------------------------------------------------
{ Abstract:
{    calls the microcode to write a sequence of buffers to 
{    streamer.
{ Parameters:
{    BufCnt - locations 1 through BufCnt of BufLocLoc contain
{             valid pointers to buffers.
{    BufLocLoc - an array of pointers to buffers.
{             The low bound of this array must be 0.
{    Start  - where in the array to start putting out.
{ Error:
{    Raises TSError if an exception is detected on tape.
{ Environment:
{    CallWrSetup must have been called before this routine.
{--------------------------------------------------------}
var
   T: Boolean;
   Status: StatBuf;
begin
   LoadAdr(Status);

   LoadAdr(BufLocLoc);
            {Compute   address of the Start'th location in BufLocLoc by
            adding an offset to (low order word of) BufLocLoc.
            The offset is for the Start elements that precede the
            desired element.  Each element is a ptrDiskBuffer.}
   LoadExpr(Start*WordSize(ptrDiskBuffer));  {Compute offset}
   InLineByte(72 {ADI});                         {add it to BufLocLoc}

   LoadExpr(BufCnt);
   LoadExpr(LOR (Shift(TSWrBuffers, 8), Shift(TsWrBuffers, -8)));
   InLineByte(229 {TLATE3});  {TLATE BufLocLoc}
   InLineByte(#277 {JCS});
   StorExpr(T);
   if not T then raise TSError(Status, 'Write Buffers failed');
end;

procedure  CallWrEnd(var Status: StatBuf);
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine that finishes a write,
{    turning streamer offline and reading in the final status.
{    (Writes threee file marks on tape.)
{ Environment:
{    Assumes the tape is near BOT or is in a write command.
{ Side Effect:
{    Leaves the tape rewinding.
{ Errors:
{    Raises TSError if any error occurs.
{--------------------------------------------------------}
var
   T: Boolean;
begin
   LoadAdr(Status);
   LoadExpr(LOR (Shift(TSWrEnd, 8), Shift(TsWrEnd, -8)));
   InLineByte(228 {TLATE2});
   InLineByte(#277 {JCS});
   StorExpr(T);
   if not T then raise TSError(Status, 'Failed trying to write end of tape');
end;


procedure  CallWrEof;
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine that writes a file mark.
{ Environment:
{    Assumes the tape is near BOT or is in a write command.
{ Error:
{    Raises TSError if an exception is detected on tape.
{--------------------------------------------------------}
var
   T: Boolean;
   Status: StatBuf;
begin
   SendCmd(#140{WrFM}, NoChange, Status, T);
   if not T then raise TSError(Status, 'Failed trying to write file mark.');
end;

procedure  CallRdSetup(var Status: StatBuf);
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine to initialize streamer
{    and set up a read command.
{ Parameter:
{    Status - The status after initializing streamer is
{        stored in this array.
{ Environment:
{    Assumes tape is at BOT - if not, gets TimeOut error.
{ Side Effects:
{    Streamer is left ONLINE and with a read command.
{--------------------------------------------------------}
var 
   T: Boolean;
begin
   SoftErrors := 0;
   OverRuns := 0;
   CallTSPosn(#041{rewind}, Status);
   LoadAdr(Status);
   LoadExpr(LOR (Shift(TSRdSetup, 8), Shift(TSRdSetup, -8)));
   InLineByte(228 {TLATE2});
   InLineByte(#277 {JCS});
   StorExpr(T);
end;

function   CallRdBuffers(var BufCnt: integer;
                         var BufLocLoc: BufList; Start: integer): Boolean;
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode to read a sequence of buffers from 
{    streamer.
{ Parameters:
{    BufCnt - locations 1 through BufCnt of BufLocLoc contain
{             valid pointers to buffers.
{         After completion, this paramter is changed to show how many 
{         buffers were read.
{    BufLocLoc - an array of pointers to buffers.
{             The low bound of this array must be 0.
{    Start     - offset of first buffer to use within array
{ Returns:
{    True if read completed okay, FALSE if file mark detected.
{    Also returns False if TSError handler returns.
{ Error:
{    Raises TSError if an exception (other than file mark)
{    is detected on tape.
{ Environment:
{    CallRdSetup must have been called before this routine.
{--------------------------------------------------------}
var
   T: Boolean;
   Status: StatBuf;
begin
   LoadAdr(Status);
   
   LoadAdr(BufLocLoc);  
              {Use as first buffer the one at BufLocLoc[Start].
              To do this, add Start*WordSize(element of BufLocLoc)
              to the BufLocLoc value.}
   LoadExpr(Start*WordSize(ptrDiskBuffer));    {get offset}
   InLineByte(72 {ADI});                           { & and add to BufLocLoc}
   
   LoadExpr(BufCnt);
   LoadExpr(LOR (Shift(TSRdBuffers, 8), Shift(TsRdBuffers, -8)));
   InLineByte(229 {TLATE3});  {TLATE BufLocLoc}
   InLineByte(#277 {JCS});
   StorExpr(T);
   BufCnt := Status[7];
   if not T then begin
      if (Land(Status[1],#357{ignore WrProt}) = #201{file mark}) then begin   
         SendCmd(#200 {Read}, NoChange, Status, T);
         SoftErrors := SoftErrors + Shift(Status[3],8) + Status[4];
         OverRuns   := OverRuns   + Shift(Status[5],8) + Status[6];
      end
      else raise TSError(Status, 'Failed in reading buffers.');
      CallRdBuffers := False;
   end
   else   
      CallRdBuffers := True;

end;

procedure  CallRdEnd(var Status: StatBuf);
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine that finishes a read,
{    turning streamer offline and reading in the final status.
{ Parameters:
{    Status - set to final status.
{ Side Effects:
{    Leaves streamer offline.  If it was in a read or write, it is rewinding
{ Errors: 
{    If status cannot be read, raises TSError
{--------------------------------------------------------}
var 
   T: Boolean;
begin
   SendCmd(#041 {Rewind}, OffLine, Status, T);
   
   Status[1] := 0;
   Status[2] := 0;
   Status[3] := Shift(SoftErrors, -8);
   Status[4] := Land (SoftErrors, #277);
   Status[5] := Shift(OverRuns, -8);
   Status[6] := Land (OverRuns, #277);
end;

procedure  CallRdEof;
{---------------------------------------------------------
{ Abstract:
{    Calls the microcode routine that reads a file mark.
{ Error:
{    Checks for exception.  Expects and ignores a file mark exception;
{    raises TSError for any other exception.
{    No distinction is made if no exception is reported.
{ Note:
{    Didn't seem able to make this work.
{--------------------------------------------------------}
var
   T: Boolean;
   Status: StatBuf;
begin
   SendCmd(#240 {ReadFM}, NoChange, Status, T);
   if not T then begin
      if (Status[1] = #201) and (Status[2]=0) then 
         SendCmd(#200{Read}, NoChange, Status, T)    {okay, continue reading}
      else raise TSError(Status, 'Failed trying to read file mark.');
   end
   else
      {debug:} write ('Success in ReadFM(!?)');
end.
