module STBufRead;

{-------------------------------------------------------
{ Copyright (C), 1981  Three Rivers Computer Corporation
{      by W. J. Hansen
{
{ Abstract:
{    This module manages buffers for reading from the streamer.
{
{ Design:
{    The buffers are initially allocated by a call on InitSTBuffers.
{    The tape must be initialized with TSInit and CallRdSetup
{-------------------------------------------------------}

{-------------------------------------------------------
{ Change Log:
{      9 Sep 81   WJH  V0.3   fix EndChunk so FileMark BitMap isn't endtape
{
{      1 Sep 81   WJH  V0.2   Don't call RdEnd in FillBuffers
{                             Put in ReadToFM
{
{     28 Aug 81   WJH  V0.1   Fix setting x-Granted in ReadBlock
{
{     19 Aug 81   WJH  V0.0   Convert from STBufWrite
{                             write FillBuffers and ReadBlock
{
{     12 Aug 81   WJH  V1.0   Convert to use STBufUts
{
{     11 Aug 81   WJH  V0.1   convert to be StBufWrite.Pas
{
{     13 July 81  WJH  V0.0   create file StBuffers
{-------------------------------------------------------}

EXPORTS

imports STBufUts from STBufUts;
imports DiskIO from DiskIO;  {for ptrDiskBuffer}

function ReadBlock(var Buf: ptrDiskBuffer): EofFlagType;
procedure ReadToFM;

PRIVATE
imports TSInterface from TSInterface;


procedure FillBuffers;
{-------------------------------------------------------
{ Abstract:
{    Reads from tape, fills buffers, and sets up Chunk table.
{-------------------------------------------------------}
var
   Filling: Boolean;
   CurrGroup: 0..MaxGroups;
   CurrBuffer: BufRange;
   Got: integer;

   procedure EndChunk(Last: BufRange;  EofFlag: EofFlagType);
   {-------------------------------
   { Abstract:
   {    Terminates one chunk and starts the next.
   { Note:
   {    A file mark followed by either a file mark or a pseudo file mark
   {    will end the reading.  (The pseudo here is to get around the
   {    problem that was occurring where the streamer was repeating 
   {    after the file mark the same block that occurred before it.
   {-------------------------------}
   var
      Status: StatBuf;
   begin
      with ChunkTable[NumChunks] do begin
         ChLast := Last;
         ChEof  := EofFlag;
         if PrevEof = EndTape then 
            ChEof := EndTape
         else if PrevEof >= FileMark then
            if EofFlag >= PseudoEof then  {This is the end of tape if no blocks
                                        have occurred since last FileMark.}
               if ChFirst > Last then begin
                  Filling := False;
                  ChEof := EndTape;
               end;
         PrevEof := ChEof;
      end;
      NumChunks := NumChunks+1;
      ChunkTable[NumChunks].ChFirst := Last+1;
      CurrBuffer := Last+1;
   end;
   
   function IsPseudo(Buf: ptrDiskBuffer): Boolean;
   begin
      IsPseudo := (Buf^.IntData[0] = ValFlag) and
                  (Buf^.IntData[1] = ValPseudo);
   end;
   
   function IsQuote(Buf: ptrDiskBuffer): Boolean;
   begin
      IsQuote := (Buf^.IntData[0] = ValFlag) and
                  (Buf^.IntData[1] = ValQuote);
   end;
   
begin
   NumChunks := 1;
   CurrBuffer := 1;
   ChunkTable[NumChunks].ChFirst := CurrBuffer;
   CurrGroup := 1;
   Filling := True;
   while Filling do  begin
      Got := 1;
      if not CallRdBuffers(Got, BufPtrs, CurrBuffer) then
         EndChunk(CurrBuffer+Got-1, FileMark)
      else if Got<1 then
         Raise TSBufError(6)
      else if IsPseudo(BufPtrs[CurrBuffer]) then
         EndChunk(CurrBuffer-1, PseudoEof)
      else begin
         if IsQuote(BufPtrs[CurrBuffer]) then begin
            Got := 1;
            if CallRdBuffers(Got, BufPtrs, CurrBuffer) then {musn't fail};
            if Got< 1 then
               Raise TSBufError(7);
         end;
         if CurrBuffer>=GroupTable[CurrGroup] then begin
            EndChunk(CurrBuffer, NotEof);
            if CurrGroup>=NumGroups then 
               Filling := False
            else CurrGroup := CurrGroup+1;
         end
         else CurrBuffer := CurrBuffer+1;
      end;
   end;
   CurrChunk := 1;
   NumChunks := NumChunks-1;
   FirstGranted := ChunkTable[CurrChunk].ChFirst;
   LastGranted  := ChunkTable[CurrChunk].ChLast;
end;

function ReadBlock(var Buf: ptrDiskBuffer): EofFlagType;
{-------------------------------------------------------
{ Abstract:
{    Gets blocks from streamer.
{ Parameters:
{    Buf - is set to point to a block
{ Returns:
{    Indication of whether there was an end of file.
{-------------------------------------------------------}
begin
   ReadBlock := NotEof;
   if CurrChunk > NumChunks then 
      FillBuffers;
   while (FirstGranted > LastGranted) and
         (ChunkTable[CurrChunk].ChEof=NotEof) do begin
      CurrChunk := CurrChunk+1;
      if CurrChunk <= NumChunks then begin
         FirstGranted := ChunkTable[CurrChunk].ChFirst;
         LastGranted  := ChunkTable[CurrChunk].ChLast;
      end
      else FillBuffers;
   end;
   if FirstGranted > LastGranted then begin  {must be some form of endfile}
      ReadBlock := ChunkTable[CurrChunk].ChEof;
      if ChunkTable[CurrChunk].ChEof<> EndTape then begin
         CurrChunk := CurrChunk+1;
         FirstGranted := ChunkTable[CurrChunk].ChFirst;
         LastGranted  := ChunkTable[CurrChunk].ChLast;
      end;
   end
   else begin
      Buf := BufPtrs[FirstGranted];
      FirstGranted := FirstGranted+1;
   end;
end;


procedure ReadToFM;
{-------------------------------------------------------
{ Abstract:
{    Advances streamer tape to just after next filemark.
{ Errors:
{    Can raise TSError from the streamer routines.
{-------------------------------------------------------}
var
   Dummy: ptrDiskBuffer;
begin
   while ReadBlock(Dummy) <> FileMark do {nothing};
end.
