module STBufUts;
{-------------------------------------------------------
{ Streamer Buffer Utilities
{
{ Copyright (C), 1981  Three Rivers Computer Corporation
{      by W. J. Hansen
{
{ Abstract:
{    This module provides buffer managment for the
{    tape streamer package.
{ Design:
{    The buffers are initially allocated from POS by
{    a call on InitSTBuffers.  They are later released with 
{    RelSTBuffers.
{
{    Buffers are allocated in groups so they can be issued for
{    consecutive memory to be given to Unit_IO.  
{    BufPtrs is an array of pointers to the buffers.
{    GroupTable[i] gives the index of the element in BufPtrs
{    for the last buffer in the group.  (GroupTable doesn't
{    change after initialization.)
{       When buffers are used, the sequence is recorded in the 
{    ChunkTable.  Each entry corresponds to consecutive buffers
{    that are part of a file.  The ChEof field indicates the type of
{    end of file, if any, to follow a chunk.
{-------------------------------------------------------}

{-------------------------------------------------------
{ Change Log:
{      8 Sep 81   WJH  V1.4   add BufferSize parameter to InitSTBuffers
{
{      1 Sep 81   WJH  V1.4   add BitMapFlag and ValBitMap
{
{     28 Aug 81   WJH  V1.3   add RelSTBuffers
{
{     19 Aug 81   WJH  V1.2   change chunk table for PseudoEof
{                             add EndTape and PrevEof
{
{     14 Aug 81   WJH  V1.1   change to use 120 blocks for second group
{                             remove TSError handler from FlushBuffers
{
{     12 Aug 81   WJH  V1.0   Create STBufUts from STBufWrite
{
{     11 Aug 81   WJH  V0.1   convert to be StBufWrite.Pas
{
{     13 July 81  WJH  V0.0   create file
{-------------------------------------------------------}

EXPORTS

imports DiskIO from DiskIO;  {for ptrDiskBuffer}

const
   MaxBuffers = 255;     {Maximum number of buffers; size of BufPtrs}
   MaxGroups  = 2;       {Maximum number of groups of consecutive buffers}
   MaxChunks  = 255;     {Maximum number of pieces of files in buffers}

   ValFlag      = #103733;   {Random value for first word of flag block}
   ValPseudo    = 0;         {second word of pseudo eof block}
   ValQuote     = 7;         {second word of quote block}
   ValBitMap    = 56;        {second word of a BitMap header block}
   
type
   BufRange = 0 .. MaxBuffers;
   BufList = array [0..MaxBuffers] of ptrDiskBuffer;
   GroupList= array [0..MaxGroups] of BufRange;

   EofFlagType = (NotEof, BitMapFlag, PseudoEof, FileMark, EndTape);
   Chunk    = record
               ChFirst, ChLast:  BufRange;
               ChEof: EofFlagType;
            end;
   ChunkList= array [1..MaxChunks] of Chunk;


var
   BufPtrs:  BufList;     {Pointers to all buffers.
                  BufPtrs[0] points to a special buffer for writing 
                  pseudo EOF blocks and Quote blocks.}
   NumBuffers: BufRange;  {last existing buffer is from BufPtrs[NumBuffers]}

   
   GroupTable: GroupList; {Delimit sections of BufPtrs that are
                          allocated consecutively}
   NumGroups: 0..MaxGroups;   {Number of consecutive groups of buffers}
          {Consecutive buffers are:
              BufPtrs[1] up to  BufPtrs[GroupTable[1]],
              BufPtrs[GroupTable[1]+1] up to BufPtrs[GroupTable[2]]
          and so on with the last ending with  BufPtrs[GroupTable[NumGroups]]}


   ChunkTable: ChunkList; {Dynamically delimit pieces of file
                          as they are inserted in the buffers}
         {The i'th chunk is buffers  BufPtrs[ChunkList[i].ChFirst]
         through   BufPtrs[ChunkList[i].ChLast].  The ChFlag says
         whether there is a PseudoEof or real FileMark after the
         chunk.  If ChLast<ChFirst, the chunk has no buffers (but may
         have an end of file).}
 
 
   FirstGranted, LastGranted: BufRange;   {Control use of buffers.
                   If LastGranted is one less than FirstGranted,
                   then none are in use and FirstGranted shows 
                   next to use.  If LastGranted is greater or equal
                   FirstGranted, then those from FirstGranted to
                   LastGranted have been allocated.}
   NumChunks: integer;
                   {Number of Chunks in use.  Initially zero.}
   CurrChunk: integer;
                   {Chunk currently in use for input.}
   
   PrevEof: EofFlagType;
                   {Eof condition after previous buffer read}

Exception TSBufError(Code: integer);   
{--------------------------------------------------------
{ Abstract: 
{    This exception is raised for program errors in calling 
{    this module.
{ Error codes:
{  for writing:
{    0 - Got less than three buffers (signalled by STFOWrite)
{    1 - The previous allocation was not done or flushed prior
{        to this call.
{    2 - There was no allocation prior to this call on done.
{    3 - The number claimed done is greater than allocated.
{    4 - FirstGranted got above NumBuffers.
{  for reading:
{    5 - STUT got filemark reading header.
{    6, 7 - STBufRead got filemark when data expected.
{-------------------------------------------------------}

procedure InitSTBuffers(BufferSize: integer);
procedure RelSTBuffers;

PRIVATE

imports Memory from Memory;
imports Screen from Screen;
imports Virtual from Virtual;


var
   GroupSegment: array [0..MaxGroups] of SegmentNumber;
           {Save segment numbers so they can be released.}


procedure InitSegment (Size: integer);
{--------------------------------------------------------
{ Abstract:
{    Initialize a segment by creating it and putting pointers to
{    its blocks into BufPtrs and GroupTable.
{ Parameters:
{    Size - number of blocks in the segment.
{--------------------------------------------------------}
var
   i : BufRange;               {for-loop index}
   BufSeg: SegmentNumber;      {id of new segment}
begin
   CreateSegment(BufSeg, Size, 1, Size);
   SetMobility(BufSeg, UnSwappable);
   SetMobility(BufSeg, UnMovable);
   for i := 1 to Size do
      BufPtrs[i+NumBuffers] := MakePtr(BufSeg, 256*(i-1), ptrDiskBuffer);
   NumBuffers := NumBuffers + Size;
   NumGroups := NumGroups+1;
   GroupTable[Numgroups] := NumBuffers;
   GroupSegment[NumGroups] := BufSeg;
end;


procedure InitSTBuffers(BufferSize: integer);
{-------------------------------------------------------
{ Abstract:
{    Gets space for buffers and stores pointers to them.
{ Side Effects:
{    BufPtrs  - initialized with pointers to pieces of buffers.
{    FirstGranted - Set to first to allocate.
{    LastGranted  - set to indicate that none have been granted.
{ Environment:
{    When the screen is shrunk to be 256 lines, 144 blocks are
{    released.  The maximum group size is 127 (for Unit_IO), so the
{    screen space allocated is 127 blocks.
{-------------------------------------------------------}
begin
   SSetSize (2*128, False, False);      {Use only the top of the screen}
   NumGroups := 0;
   GroupTable[0] := 0;
   NumBuffers := 0;
   new(BufPtrs[0]);
   BufPtrs[0]^.intdata[0] := ValFlag;   {set up special flag block buffer}
   InitSegment(127);                {Alloc most of screen for first group}
   if BufferSize>0 then
      InitSegment(BufferSize);
   FirstGranted := 1;
   LastGranted := 0;
   NumChunks := 0;
   CurrChunk := 1;
   PrevEof := NotEof;
end;


procedure RelSTBuffers;
{-----------------------
{ Abstract:
{    Releases buffers allocated by InitStBuffers.
{ Design:
{    Required only if calling program is to continue processing.
{    The screen is not reset by this function because the system
{    doesn't provide a way to do it.  See the code for a number
{    of approaches which don't work.
{-----------------------}
var 
   i: 1..MaxGroups;
begin
   for i := 1 to NumGroups do begin
      SetMobility(GroupSegment[i], Swappable);
          {Need to set mobility because operating system thinks its 
          unmovable even though unused.}
      DecRefCount(GroupSegment[i]);
   end;
{  SFullWindow;    }
{SFullWindow doesn't seem to make it use the full screen.}
{   SSetSize(8*128, False, False);    }
{Can't SSetSize because the code gets loaded into the screen area.}
           {Even if I release segments in reverse order, something necessary
           to move segments gets loaded into the screen area.}
end.
