module STBufWrite;

{-------------------------------------------------------
{ Copyright (C), 1981  Three Rivers Computer Corporation
{      by W. J. Hansen
{
{ Abstract:
{    This module manages buffers for writing to the streamer.
{
{ Design:
{    The buffers are initially allocated by a call on InitBuffers.
{
{    Thereafter, for writing, the client program asks for
{    a region of consecutive memory by calling AllocWrBuffers.
{    These are actually written sometime after a call
{    to DoneWrBuffers.  The argument to the latter says
{    how many were actually used.
{    A call to FlushWrBuffers puts all used buffers to 
{    tape, even though all buffers may not be filled.
{
{    After AllocWrBuffers, FirstGranted and LastGranted point
{    to the first and last buffers granted by AllocWrBuffers.
{    After DoneWrBuffers (and initially), FirstGranted points
{    to next to use and LastGranted is one lower.
{-------------------------------------------------------}

{-------------------------------------------------------
{ Change Log:
{      1 Sep 81   WJH  V1.4   write quote records before those with ValFlag
{                             add BitMapFlag processing in FlushBuffers
{
{     21 Aug 81   WJH  V1.3   allow 0 blocks for alloc and done
{
{     20 Aug 81   WJH  V1.2   add processing for EndTape
{
{     19 Aug 81   WJH  V1.1   add PseudoEof
{
{     12 Aug 81   WJH  V1.0   Convert to use STBufUts
{
{     11 Aug 81   WJH  V0.1   convert to be StBufWrite.Pas
{
{     13 July 81  WJH  V0.0   create file
{-------------------------------------------------------}

EXPORTS

imports STBufUts from STBufUts;
imports DiskIO from DiskIO;  {for ptrDiskBuffer}

function  AllocWrBuffers(NumRequested: BufRange; var NumGiven: BufRange):
                        BufRange;
procedure DoneWrBuffers(NumDone: BufRange;  EofFlag: EofFlagType);
procedure FlushWrBuffers;

PRIVATE
imports TSInterface from TSInterface;


function  AllocWrBuffers(NumRequested: BufRange; var NumGiven: BufRange):
                        BufRange;
{-------------------------------------------------------
{ Abstract:
{    Allocates consecutive buffers.
{ Parameters:
{    NumRequested - Number of buffers requested.
{    NumGiven     - Number actually given.  May be less
{       than NumRequested due to reaching the end of the 
{       NumBuffers buffers in the pool.
{ Returns:
{    subscript (into BufPtrs) of first of the consecutive buffers.
{ Side Effects:
{    FirstGranted - set to index of first buffer granted.
{    LastGranted - Set to index of last buffer granted.
{ Error:
{    Raises TSBufError if the previous alloc has not been
{    completed with DoneWriteBuffers.
{-------------------------------------------------------}
var
   Group: 0..MaxGroups;
begin
   if FirstGranted <= LastGranted then
      raise TSBufError(1);
   if FirstGranted > NumBuffers then 
      raise TSBufError(4);
   AllocWrBuffers := FirstGranted;

   Group := 1;                     {determine group}
   while FirstGranted > GroupTable[Group] do
      Group := Group+1;

   {LastGranted is the smaller of number requested and distance to
                    end of group.} 
   if FirstGranted+NumRequested-1 > GroupTable[Group] then
      LastGranted := GroupTable[Group]
   else
      LastGranted := FirstGranted+NumRequested-1;   
   NumGiven := LastGranted+1 - FirstGranted;
end;

procedure DoneWrBuffers(NumDone: BufRange;  EofFlag: EofFlagType);
{-------------------------------------------------------
{ Abstract:
{    Deallocates consecutive buffers.  Makes entries in 
{    ChunkTable to describe the chunk.  Flushes buffers
{    if buffer pool is full.
{ Parameters:
{    NumDone - Number of buffers filled.
{        Full buffers are those in range  FirstGranted ...
{                FirstGranted+NumDone-1.
{    EofFlag - Value is stored as the chunk ChEof flag.
{ Errors:
{    Raises TSBufError if there has been no prior Alloc.
{ Side Effects:
{    FirstGranted - set to last used.
{    LastGranted - set to one before last used.
{-------------------------------------------------------}
var
   LastUsed: BufRange;
begin
   LastUsed := FirstGranted+NumDone-1;
   if LastUsed > LastGranted then
      raise TSBufError(3);
   if (FirstGranted>LastGranted) and (NumDone>0) then
      raise TSBufError(2);
   NumChunks := NumChunks +1;
   with ChunkTable[NumChunks] do begin
      ChFirst := FirstGranted;
      ChLast  := LastUsed;
      ChEof   := EofFlag;
   end;
   if (LastUsed >= NumBuffers) or (NumChunks+2 > MaxChunks) then 
      FlushWrBuffers
   else begin
      FirstGranted := LastUsed+1;
      LastGranted := LastUsed;
   end;
end;

procedure FlushWrBuffers;
{-------------------------------------------------------
{ Abstract:
{    Writes the buffer pool to tape, even if it is not full.
{ Side Effects:
{    FirstGranted - set to allocated first buffer.
{    LastGranted - set to indicate none granted.
{-------------------------------------------------------}
var
   i: 0..MaxChunks;
   Status: StatBuf;
   j: BufRange;
begin
   for i := 1 to NumChunks do with ChunkTable[i] do begin
      j := ChFirst;
      while j <= ChLast do begin      {check for blocks that need quoting}
         if BufPtrs[j]^.IntData[0] = ValFlag then begin
            if j > ChFirst then    {write prior blocks}
               CallWrBuffers(j-CHFirst, BufPtrs, ChFirst);
            BufPtrs[0]^.IntData[1] := ValQuote;
            CallWrBuffers(1, BufPtrs, 0);    {write quote block}
            ChFirst := j;            {quoted block will start next bunch}
         end;
         j := j+1;
      end;
      if ChFirst <= ChLast then   {write final part of chunk}
         CallWrBuffers(ChLast-ChFirst+1,       {the number of buffers}
                       BufPtrs, ChFirst);      {location of first}
      case ChEof of
         BitMapFlag: begin
            BufPtrs[0]^.intdata[1] := ValBitMap;
            CallWrBuffers(1, BufPtrs, 0);
         end;
         PseudoEof: begin
            BufPtrs[0]^.intdata[1] := ValPseudo;
            CallWrBuffers(1, BufPtrs, 0);
         end;
         FileMark:
            CallWrEof;
         EndTape: begin
            CallWrEnd(Status);
            PrtStatus(Status, 'Final statistics');
         end;
      end {case};
   end;
   FirstGranted := 1;
   LastGranted := 0;
   NumChunks := 0;
end.
