module STFORead;      
{---------------------------------------------------
{ Abstract:
{    Reads files from streamer 
{    and loads them onto disk or prints their names.
{
{    by W. J. Hansen
{    copyright 1981, Three Rivers Computer Corporation
{---------------------------------------------------}

{---------------------------------------------------
{ Change Log:
{
{   28 Oct 81  WJH V0.5   Fix bug in creating directories
{
{   26 Oct  81 WJH V0.4   use upper case of tape partition names.
{                         EnterFile changed to create any needed directories
{                         put in Virgil headers for procedures
{                         reinstate the check that file is newly created.
{
{    3 Sept 81 WJH V0.3   check that file has been created
{
{    2 Sept 81 WJH V0.2   print names in two columns
{                         count files and blocks
{
{    1 Sept 81 WJH V0.1   put in code to write to disk
{
{    21 Aug 81 WJH V0.0   create program that just prints file names
{--------------------------------------------------}
exports

imports Stut from Stut;      {for ptrSelectEntry}
   
procedure FORead(Selections: ptrSelectEntry; BufferSize: integer);
procedure TapeDirectory(BufferSize: integer);

   
private                        

   imports TSInterface from TSInterface;
   imports STBufUts    from STBufUts;        
   imports STBufRead   from STBufRead;        

   imports FileDefs    from FileDefs;        
   imports FIleSystem  from FileSystem;
   imports FileUtils   from FileUtils;
   imports DiskIO      from DiskIO;        
   imports Clock       from Clock;
   imports Perq_String from Perq_String;        

var 
   TotalBlocks, TotalFiles, TotalErrors: integer;
                           {Totals for data in partition.}
   GrandTBlocks, GrandTFiles, GrandTErrors: integer;
                           {Grand totals}
   EndFlag: EofFlagType;

   pFSData: ptrFSDataEntry;   {global so alloc only once}
   Half: (Left, Right);
   


function EnterFile (Path: PathName;  Name: PartialPathName): FileId;
{-----------------------------------
{ Abstract:
{    Enters file with given name in directory given by path.
{    If necessary, creates subordinate directories as indicated by Name.
{ Parameters:
{    Path - Name of a directory which exists.  Name ends with '>'.
{    Name - File name to be entered.
{ Returns:
{    location of FIB for new file.
{----------------------------------}
var
   FileName: PathName;
 
   procedure MakeADir(TFileName: PathName);
   {-------------------------------------
   { Abstract: 
   {   Make a directory for last directory name in FileName.
   { Design: 
   {   Tries to make a directory for the item before the last '>' in
   {   TFileName.  If this fails, recur to make higher level directories.
   {   If the name is reduced to having no '>' then the directory 
   {   cannot be made  (see documentation of FSMakeDirectory.
   {   The error message uses the variable FileName (from outer routine
   {   in reporting the error.
   {   Similarly, the MkDirErr handler uses the DirName variable from 
   {   MakeADir to see what directory to create.
   {------------------------------------}
   label 1;
   var
      DirName: PathName;
      GTLoc:   integer;
      Dir:     FileId;

      handler MkDirErr(Msg: string;  TName: PathName);
      begin
         MakeADir(DirName);
         MakeADir(TFileName);
         goto 1;
      end;
   begin   {MakeADir}
      GTLoc := RevPosC(TFileName, '>');
      if GTLoc = 0 then begin
         LogError (concat('Unable to make directories for file', FileName))
      end
      else begin
         DirName := SubStr (TFileName, 1, GTLoc-1);
         Dir := FSMakeDirectory(DirName);
      end;
   1:
   end;   {MakeADir}
         
var
   FileLoc: FileID;
   L: integer;
   Blank: String[41];
begin   {EnterFile}
   if length(Path)+length(Name)>100 then  begin
      FileName := Path;
      AppendString (FileName, SubStr(Name, 1, 100-length(Path)));
      LogError(concat('Name truncated to 100 chars: ', FileName));
   end
   else
      FileName := concat(Path, Name);
   FileLoc := FSEnter(FileName);
   if FileLoc = 0 then begin
      MakeADir(FileName);
      FileLoc := FSEnter(FileName);
   end;
   if Half = Left then begin
      write(FileName);
      L := length(FileName);
      if L > 40 then 
         writeln
      else begin
         Blank:=substr('                                         ',1,41-L);
         write(Blank);
         Half := Right;
      end;
   end
   else begin
      writeln(FileName);
      Half := Left;
   end;
   if FileLoc=0 then 
      LogError('Could not create file')
   else begin 
      FSGetFSData(FileLoc, pFSData);   {check that it is new}
      with pFSData^ do   
         if    (FileCreateDate.Second <> FileAccessDate.Second)
            or (FileCreateDate.Minute <> FileAccessDate.Minute)
            or (FileCreateDate.Hour   <> FileAccessDate.Hour)
            or (FileCreateDate.Day    <> FileAccessDate.Day)
            or (FileCreateDate.Month  <> FileAccessDate.Month)
            or (FileCreateDate.Year   <> FileAccessDate.Year) then {not new}
          LogError('Replacing existing file');
   end;
   EnterFile := FileLoc;
end;   {EnterFile}


procedure GetPartition(Dir: DiskAddr;   Path: PathName);
{-----------------------------------
{ Abstract:
{   Reads the files from a tape partition and writes them on disk.
{
{ Parameters:
{   Dir - Disk location of directory to put files into
{   Path - String name of the directory in Dir.  Must have final '>'.
{ Environment:
{   Must have had TSInit, InitSTBuffers, and CallRdSetup.
{   Assumes ReadBlock will next return the FIB of the first file.
{ Side Effects:
{   Repositions ReadBlock to just after the FileMark at the end of
{   the tape partition.
{-----------------------------------}
var
   Dest: FileId;
   FileIdBlk: ptrDiskBuffer;
   Data: ptrDiskBuffer;
   LogBlkNum: integer;
   SaveFSData: FSDataEntry;
begin
   Half := Left;
   while ReadBlock(FileIdBlk) <> FileMark do begin        {process a file}
      SaveFSData := FileIdBlk^.FSData;
      Dest := EnterFile(Path, SaveFSData.FileName);   {enter it}
      if Dest<>0 then begin
         TotalFiles := TotalFiles+1;
         
         LogBlkNum := 0;
         EndFlag := ReadBlock(Data);
         while EndFlag=NotEof do begin       {write blocks of file}
            FSBlkWrite(Dest, LogBlkNum, recast(Data, PDirBlk));
            TotalBlocks := TotalBlocks+1;
            LogBlkNum := LogBlkNum+1;
            EndFlag := ReadBlock(Data);
         end;
         FSClose(Dest, LogBlkNum, SaveFSData.FileBits);
         
         FSGetFSData(Dest, pFSData);                      {replace new FSData}
         SaveFSData.FileName := pFSData^.FileName;        {other than name   }
         SaveFSData.FileBlocks := pFSData^.FileBlocks;    {and # blocks      }
         pFSData^ := SaveFSData;                          {with the old FSData}
         FSSetFSData(Dest, pFSData);
    
         if EndFlag=BitMapFlag then  begin
            LogError (concat('Sparse file not renumbered: ',
                          pFSData^.FileName));
            while ReadBlock(Data)<PseudoEof do {nothing};
         end;
      end
      else 
         while ReadBlock(Data)<PseudoEof do {nothing};
   end;
   if Half = Right then
      writeln;
end;
         


procedure FORead(Selections: ptrSelectEntry; BufferSize: integer);
{--------------------------------------------------
{ Abstract:
{   Read files from tape to disk in file order.
{ Parameters:
{   Selections - a ptr to a list of selected partition names
{   BufferSize - number of blocks of buffer to use from main
{       memory (in addition to 120 blocks from screen)
{       100 works fine if swapping is allowed, but no more than
{       20 should be used when booting from floppy.
{--------------------------------------------------}
label 30;
handler TSError(Status:StatBuf; Msg: TSMsg);
begin
   PrtStatus(Status, Msg);
   goto 30;
end;

var
   Start, Now: TimeString;         
   Status: StatBuf;
   BufPtr, Junk: ptrDiskBuffer;
   PartHdr: ptrPartBlk;
   Found: Boolean;
   TSel: ptrSelectEntry;
   TpPtName: string;
   NSels: integer;
begin
   GetTString(Start);  Writeln(Start);
   InitSTBuffers(BufferSize);
   CallRdSetup(Status);
   PrtStatus(Status, 'Initial status before reading');
   ReadToFM;    {skip tape header}
   GrandTBlocks := 0;
   GrandTFiles  := 0;
   GrandTErrors := 0;
   new (pFSData);

   NSels := 0;
   TSel := Selections;
   while TSel <> Nil do begin
      NSels := NSels+1;
      TSel := TSel^.SelNext;
   end;
   EndFlag := ReadBlock(BufPtr);      {read 1st partition header}
   EndFlag := ReadBlock(BufPtr);      {read second copy}
   while (EndFlag <> EndTape) and (NSels>0) do begin  {process tape partition}
      TotalBlocks := 0;  TotalFiles := 0;  TotalErrors := 0;
      EndFlag := ReadBlock(BufPtr);   {use the third copy of header}
      PartHdr := recast (BufPtr, ptrPartBlk);
      while ReadBlock(Junk) <> PseudoEof do {nothing};
      
      TSel := Selections;
      Found := False;
      while TSel <> Nil do begin
         TpPtName := PartHdr^.PartName;
         ConvUpper(TpPtName);
         if TpPtName <> TSel^.SelTpPart then 
            TSel := TSel^.SelNext
         else begin
            found := True;
            writeln;
            writeln('Tape partition ', PartHdr^.PartName,
                    '     (originally ', PartHdr^.PartOriginal,')');
            FSAddToTitleLine(concat(STUTTitle, concat('   Get to ',
                       TSel^.SelDir)));
      
            GetPartition(TSel^.SelAddr, TSel^.SelDir);
      
            Writeln('   Blocks:', TotalBlocks:6,
                    '   Files:',  TotalFiles:6,
                    '   Disk Errors:', TotalErrors:4);
            GrandTBlocks := GrandTBlocks + TotalBlocks;
            GrandTFiles  := GrandTFiles  + TotalFiles;
            GrandTErrors := GrandTErrors + TotalErrors;
            TSel^.SelTpPart := '';   {don't use this selection again}
            NSels := NSels - 1;
            TSel := Nil;    {stop checking this partition; go to next}
         end;
      end;
      if not Found then ReadToFM;
      EndFlag := ReadBlock(BufPtr);         {read next partition header}
      EndFlag := ReadBlock(BufPtr);         {read second copy}
   end;
30:
   RelSTBuffers;
   CallRdEnd(Status);
   PrtStatus(Status, 'Final Status after reading');      
   GetTString(Now);  
   writeln;
   writeln('Started reading at ', Start, '   Finished at ', Now);
   Writeln('   Total Blocks:', GrandTBlocks:6,
           '   Total Files:',  GrandTFiles:6,
           '   Total Disk Errors:', GrandTErrors:4);
end;



procedure PrintFileNames;
{-----------------------------------------------
{ Abstract:
{   Scans streamer tape and prints names of files found there.
{----------------------------------------------}
var
   BufPtr: ptrDiskBuffer;
   L: integer;
   Half: (Left, Right);
   Blank: string[41];
begin
   EndFlag := ReadBlock(BufPtr);  {Read first FIB}
   Half := Left;
   while EndFlag < FileMark do begin
      if Half = Left then begin
         Write(BufPtr^.FSData.FileName);
         L := length(BufPtr^.FSData.FileName);
         if L > 40 then 
            writeln
         else begin
            Blank:=substr('                                         ',1,41-L);
            write(Blank);
            Half := Right;
         end;
      end
      else begin
         Writeln(BufPtr^.FSData.FileName);
         Half := Left;
      end;
      TotalFiles := TotalFiles + 1;
      while EndFlag<PseudoEof do begin
         TotalBlocks := TotalBlocks + 1;
         EndFlag := ReadBlock(BufPtr);
      end;
      TotalBlocks := TotalBlocks-1;   {forget the PseudoEof}
      EndFlag := ReadBlock(BufPtr);   {Read next FIB or the FileMark}
   end;
   if Half = Right then 
      writeln;
end;



procedure TapeDirectory(BufferSize: integer);
{--------------------------------------------
{ Abstract:
{   Processes all partitions on streamer tape, calling
{   PrintFileNames to print all names.
{ Parameters:
{   BufferSize - number of non-screen buffers to use.
{         (100 for swapping.  20 if boot from floppy.)
{--------------------------------------------}
label 30;
handler TSError(Status:StatBuf; Msg: TSMsg);
begin
   PrtStatus(Status, Msg);
   GrandTBlocks := GrandTBlocks + TotalBlocks;
   GrandTFiles  := GrandTFiles  + TotalFiles;
   goto 30;
end;

var
   BufPtr, Junk: ptrDiskBuffer;
   PartHdr: ptrPartBlk;
   Status: StatBuf;
begin
   InitSTBuffers(BufferSize);
   CallRdSetup(Status);
   PrtStatus(Status, 'Initial status before reading');
   ReadToFM;    {skip tape header}
   GrandTBlocks := 0;
   GrandTFiles  := 0;

   EndFlag := ReadBlock(BufPtr);      {read 1st partition header}
   EndFlag := ReadBlock(BufPtr);      {read second copy}
   while EndFlag <> EndTape do begin   {process partition on tape}
      TotalBlocks := 0;  TotalFiles := 0;
      EndFlag := ReadBlock(BufPtr);   {use the third copy of header}
      PartHdr := recast (BufPtr, ptrPartBlk);
      while ReadBlock(Junk) <> PseudoEof do {nothing};
      writeln;
      writeln('Tape partition ', PartHdr^.PartName,
              '     (originally ', PartHdr^.PartOriginal,')');
      FSAddToTitleLine(concat(STUTTitle, concat('   Tape Partition: ',
                 PartHdr^.PartName)));

      PrintFileNames;

      Writeln('   Blocks:', TotalBlocks:6,
              '   Files:',  TotalFiles:6);
      GrandTBlocks := GrandTBlocks + TotalBlocks;
      GrandTFiles  := GrandTFiles  + TotalFiles;
      EndFlag := ReadBlock(BufPtr);         {read next partition header}
      EndFlag := ReadBlock(BufPtr);         {read second copy}
   end;
30:
   RelSTBuffers;
   CallRdEnd(Status);
   PrtStatus(Status, 'Final Status after directory');      
   writeln;
   writeln('   Total Blocks:', GrandTBlocks:6,
           '   Total Files:',  GrandTFiles:6);
end.

